/**************************************************************************/
/*                                                                        */
/*  This file is part of Frama-Clang                                      */
/*                                                                        */
/*  Copyright (C) 2012-2018                                               */
/*    CEA (Commissariat à l'énergie atomique et aux énergies              */
/*         alternatives)                                                  */
/*                                                                        */
/*  you can redistribute it and/or modify it under the terms of the GNU   */
/*  Lesser General Public License as published by the Free Software       */
/*  Foundation, version 2.1.                                              */
/*                                                                        */
/*  It is distributed in the hope that it will be useful,                 */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*  GNU Lesser General Public License for more details.                   */
/*                                                                        */
/*  See the GNU Lesser General Public License version 2.1                 */
/*  for more details (enclosed in the file LICENSE).                      */
/*                                                                        */
/**************************************************************************/

//
// Description:
//   Implementation of the ACSL++ comment.
//

#include "AnnotationComment.h"
#include "ACSLComment.h"
#include "ACSLComponent.h"
#include "clang/Parse/Parser.h"
#include "clang/AST/ASTConsumer.h"
#include "clang/Sema/Scope.h"
#include <string>

ACSLComment::Kind
ACSLComment::getAnnotationKind(const char*& start, size_t length) const {
  if (start[0] != '@')
    return KUndefined;
  size_t firstIndex = 1;
  while (true) {
    while (firstIndex < length && isspace(start[firstIndex]))
      ++firstIndex;
    if (firstIndex < length-1 && start[firstIndex] == '/') {
      if (start[firstIndex+1] == '/') {
        firstIndex += 2;
        while (firstIndex < length && (start[firstIndex] != '\n'))
          ++firstIndex;
      }
      else
        break;
    }
    else
      break;
  };

  if (firstIndex >= length)
    return KUndefined;
  if (!isalpha(start[firstIndex]))
    return KUndefined;
  size_t endIndex = firstIndex+1; 
  while (endIndex < length && isalpha(start[endIndex]))
    ++endIndex;

  std::string firstWord(&start[firstIndex], (endIndex-firstIndex));
  std::transform(firstWord.begin(), firstWord.end(), firstWord.begin(),
      ::tolower);
  if (firstWord == "ghost") {
    _startOfGhost = endIndex;
    return KGhost;
  }
  if (firstWord == "requires" || firstWord == "terminates"
      || firstWord == "decreases" || firstWord == "assigns"
      || firstWord == "ensures" || firstWord == "behavior"
      || firstWord == "assumes" || firstWord == "complete"
      || firstWord == "disjoint" || firstWord == "allocates"
      || firstWord == "frees" || firstWord == "exists")
    return KNext; /* function or statement */
  if (firstWord == "for") {
    const char* lookForColon = &start[endIndex];
    size_t currentLength = length - endIndex;
    while (*lookForColon && *lookForColon != ':' && currentLength > 0) {
      ++lookForColon;
      --currentLength;
    };
    if (currentLength > 0 && *lookForColon == ':') {
      firstIndex = endIndex + (lookForColon - &start[endIndex])+1;
      while (firstIndex < length && isspace(start[firstIndex]))
        ++firstIndex;
      if (firstIndex >= length)
        return KNextStatement;
      if (!isalpha(start[firstIndex]))
        return KNextStatement;
      endIndex = firstIndex+1; 
      while (endIndex < length && isalpha(start[endIndex]))
        ++endIndex;
      firstWord = std::string(&start[firstIndex], (endIndex-firstIndex));
      std::transform(firstWord.begin(), firstWord.end(), firstWord.begin(),
          ::tolower);
    }
    else
      return KNextStatement;
  };
  if (firstWord == "loop" || firstWord == "for") {
    return KNextLoop; /* loop or statement */
  }
  if (firstWord == "assert")
    return KLabel; /* assertion */
  if (firstWord == "breaks" || firstWord == "continues"
      || firstWord == "returns")
    return KNextStatement; /* statement */
  if (firstWord == "invariant")
    return KOuterLoop;
  if (firstWord == "logic" || firstWord == "predicate" || firstWord == "global")
    return KGlobal;
  if (firstWord == "inductive" || firstWord == "axiomatic" 
      || firstWord == "lemma" || firstWord == "axiom")
    // strictly speaking, axiom cannot appear independently of an axiomatic,
    // but this error is catched in the GlobalAnnotation parser itself.
    return KGlobal;
  if (firstWord == "model")
    return KGlobal; /* declaration: variable or field */
  if (firstWord == "class" || firstWord == "type")
    return KGlobal; /* in the scope of the token */
  if (firstWord == "ghost")
    return KNext; /* KPrevious ??? declaration or statement */
  if (firstWord == "volatile")
    return KGlobal; /* declaration */
  return KUndefined;
}

void
ACSLComment::parseGlobal(ForwardReferenceList& globals,
    ForwardReferenceList* classContent, const clang::DeclContext* clangContext,
    clang::ASTContext* astContext, clang::Sema* sema,
    clang::Scope* scope, Clang_utils* clangUtils,
    const RTTITable& rttiTable, location loc) {
  Acsl::GlobalAnnotation globalAnnotation(loc);
  Acsl::Parser parser(clangContext, astContext, sema, scope,
    getSourceLocation(), clangUtils, globalAnnotation, rttiTable,
    copy_loc(loc));
  parser.parse(getContent());
  /*global_annotation*/list annots = globalAnnotation.getAnnot();
  while(annots) {
    global_annotation annot =
      (global_annotation)annots->element.container;
    if (classContent) {
      class_decl class_annot = class_decl_Class_annot(loc,annot);
      classContent->insertContainer(class_annot);
    }
    else {
      translation_unit_decl decl =
        translation_unit_decl_GlobalAnnotation(loc,annot);
      globals.insertContainer(decl);
    }
    annots = annots->next;
  }
}

void ACSLComment::parseGhostGlobal(
  //ForwardReferenceList& globals,
  //ForwardReferenceList* classContent,
    clang::DeclContext* clangContext,
    //    clang::ASTContext* astContext,
    clang::Sema* sema,
    clang::Scope* scope,
    clang::CompilerInstance& compilerInstance,
    clang::ASTConsumer* consumer) {
  llvm::StringRef code = llvm::StringRef(getContent());
  code = code.substr(_startOfGhost);
  clang::SourceLocation loc =
    getSourceLocation().getLocWithOffset(_startOfGhost);
  clang::Preprocessor& cpp = compilerInstance.getPreprocessor();
  clang::Parser* parser =
    static_cast<clang::Parser*>(cpp.getCodeCompletionHandler());
  std::unique_ptr<llvm::MemoryBuffer> buff =
    llvm::MemoryBuffer::getMemBufferCopy(code, "ghost.cxx");
  clang::SourceManager& source_manager = compilerInstance.getSourceManager();
  clang::FileID file =
    source_manager.createFileID(
	    std::move(buff),clang::SrcMgr::C_User,0,0,loc);
  cpp.EnterSourceFile(file,0,loc);
  cpp.enableIncrementalProcessing();
  clang::Parser::DeclGroupPtrTy Decl;
  while (!parser->ParseTopLevelDecl(Decl)) {
    clang::DeclGroupRef DGR = Decl.get();
    consumer->HandleTopLevelDecl(DGR);
  }
  cpp.enableIncrementalProcessing(false);
  cpp.EndSourceFile();
}

void ACSLComment::parseGhostStatement(
    clang::DeclContext* clangContext,
     clang::Sema* sema,
    clang::Scope* scope,
    clang::CompilerInstance& compilerInstance,
    clang::ASTConsumer* consumer) {
  std::cerr << "Not yet implemented, ignoring ghost statement" << std::endl;
}

void
ACSLComment::parseCodeAnnotation(ForwardReferenceList& codeContainer,
    const clang::DeclContext* clangContext, clang::ASTContext* astContext,
    clang::Sema* sema, clang::Scope* scope, Clang_utils* clangUtils,
    const RTTITable& rttiTable, location loc) {
  Acsl::CodeAnnotation codeAnnotation(loc);
  Acsl::Parser parser(clangContext, astContext, sema, scope,
    getSourceLocation(), clangUtils, codeAnnotation, rttiTable, copy_loc(loc));
  parser.parse(getContent());
  code_annotation annotation = codeAnnotation.extractAnnot();
  if (annotation)
    codeContainer.insertContainer(statement_Code_annot(
        codeAnnotation.extractLocation(), annotation));
}

/* code_annotation */ list
ACSLComment::parseLoopAnnotation(const clang::DeclContext* clangContext,
    clang::ASTContext* astContext, clang::Sema* sema, clang::Scope* scope,
    Clang_utils* clangUtils, const RTTITable& rttiTable, location loc) {
  Acsl::LoopAnnotation loopAnnotation(loc);
  Acsl::Parser parser(clangContext, astContext, sema, scope,
    getSourceLocation(), clangUtils, loopAnnotation, rttiTable, copy_loc(loc));
  parser.parse(getContent());
  return loopAnnotation.extractResult();
}

void
ACSLComment::parseStatementAnnotation(ForwardReferenceList& codeContainer,
    const clang::DeclContext* clangContext, clang::ASTContext* astContext,
    clang::Sema* sema, clang::Scope* scope, Clang_utils* clangUtils,
    const RTTITable& rttiTable, location loc) {
  Acsl::StatementAnnotation statementAnnotation(codeContainer, loc);
  Acsl::Parser parser(clangContext, astContext, sema, scope,
    getSourceLocation(), clangUtils, statementAnnotation, rttiTable,
    copy_loc(loc));
  parser.parse(getContent());
}

void
ACSLComment::parseFunctionContract(option& /* function_contract */ contract,
    const clang::DeclContext* clangContext, clang::ASTContext* astContext,
    clang::Sema* sema, clang::Scope* scope, Clang_utils* clangUtils,
    const RTTITable& rttiTable, location loc) {
  Acsl::FunctionContract functionContract(loc);
  Acsl::Parser parser(clangContext, astContext, sema, scope,
    getSourceLocation(), clangUtils, functionContract, rttiTable,
    copy_loc(loc));
  parser.parse(getContent());
  contract = opt_some_container(functionContract.getAnnot());
}

AnnotationComment*
AnnotationCommentFactory::createAnnotationComment(
      const clang::SourceManager& sourceMgr, const clang::SourceRange& range)
{ ACSLComment comment(sourceMgr, range);
  return (comment.isValid()) ? new ACSLComment(comment) : NULL;
}

